/*
 * This file is part of the Cardinal Optimizer, all rights reserved.
 */

#include "coptcpp_pch.h"

#include <cmath>

using namespace std;

void solve_soc()
{
  try
  {
    // Create COPT environment
    Envr env = Envr();

    // Create COPT model
    Model model = env.CreateModel("solve_soc");

    // Add variables
    //
    //   minimize: z
    //
    //   bnds:
    //     x, y, t free, z non-negative
    //
    Var x = model.AddVar(-COPT_INFINITY, COPT_INFINITY, 0.0, COPT_CONTINUOUS, "x");
    Var y = model.AddVar(-COPT_INFINITY, COPT_INFINITY, 0.0, COPT_CONTINUOUS, "y");
    Var z = model.AddVar(0, COPT_INFINITY, 0.0, COPT_CONTINUOUS, "z");
    Var t = model.AddVar(-COPT_INFINITY, COPT_INFINITY, 0.0, COPT_CONTINUOUS, "t");

    // Set objective: z
    model.SetObjective(z, COPT_MINIMIZE);

    // Add constraints
    //
    //   r0: 3*x + y >= 1
    //   c0: z^2 >= x^2 + 2*y^2
    //
    // c0 is converted to:
    //
    //   r1: sqrt(2.0)*y - t = 0
    //   c1: z^2 >= x^2 + t^2
    //
    model.AddConstr(3.0 * x + y >= 1.0, "r0");
    model.AddConstr(sqrt(2.0) * y - t == 0, "r1");

    VarArray cvars;
    cvars.PushBack(z);
    cvars.PushBack(x);
    cvars.PushBack(t);
    model.AddCone(cvars, COPT_CONE_QUAD);

    // Set parameters
    model.SetDblParam(COPT_DBLPARAM_TIMELIMIT, 10);

    // Solve the model
    model.Solve();

    // Analyze solution
    if (model.GetIntAttr(COPT_INTATTR_LPSTATUS) == COPT_LPSTATUS_OPTIMAL)
    {
      cout << endl;
      cout << "Objective value: " << model.GetDblAttr(COPT_DBLATTR_LPOBJVAL) << endl;

      VarArray vars = model.GetVars();
      cout << "Variable solution:" << endl;
      for (int i = 0; i < vars.Size(); ++i)
      {
        Var var = vars.GetVar(i);
        cout << "  " << var.GetName() << " = " << var.Get(COPT_DBLINFO_VALUE) << endl;
      }
      cout << endl;
    }
  }
  catch (CoptException e)
  {
    cout << "Error code = " << e.GetCode() << endl;
    cout << e.what() << endl;
  }
  catch (...)
  {
    cout << "Exception during optimization" << endl;
  }
}

void solve_rsoc()
{
  try
  {
    // Create COPT environment
    Envr env = Envr();

    // Create COPT model
    Model model = env.CreateModel("solve_rsoc");

    // Add variables
    //
    //   minimize: 1.5*x - 2*y + z
    //
    //   bnds:
    //     0 <= x <= 20
    //     y, z, r >= 0
    //     s, t free
    //
    Var x = model.AddVar(0.0, 20.0, 0.0, COPT_CONTINUOUS, "x");
    Var y = model.AddVar(0.0, COPT_INFINITY, 0.0, COPT_CONTINUOUS, "y");
    Var z = model.AddVar(0.0, COPT_INFINITY, 0.0, COPT_CONTINUOUS, "z");
    Var r = model.AddVar(0.0, COPT_INFINITY, 0.0, COPT_CONTINUOUS, "r");
    Var s = model.AddVar(-COPT_INFINITY, COPT_INFINITY, 0.0, COPT_CONTINUOUS, "s");
    Var t = model.AddVar(-COPT_INFINITY, COPT_INFINITY, 0.0, COPT_CONTINUOUS, "t");

    // Set objective: 1.5*x - 2*y + z
    model.SetObjective(1.5*x - 2*y + z, COPT_MINIMIZE);

    // Add constraints
    //
    //   r0: 2*x + y >= 2
    //   r1: -x + 2*y <= 6
    //   r2: r = 1
    //   r3: 2.8284271247 * x + 0.7071067811 * y - s = 0
    //   r4: 3.0822070014 * y - t = 0
    //   c0: 2*z*r >= s^2 + t^2
    //
    model.AddConstr(2*x + y >= 2, "r0");
    model.AddConstr(-x + 2 * y <= 6, "r1");
    model.AddConstr(r == 1, "r2");
    model.AddConstr(2.8284271247 * x + 0.7071067811 * y - s == 0, "r3");
    model.AddConstr(3.0822070014 * y - t == 0, "r4");

    VarArray rvars;
    rvars.PushBack(z);
    rvars.PushBack(r);
    rvars.PushBack(s);
    rvars.PushBack(t);
    model.AddCone(rvars, COPT_CONE_RQUAD);

    // Set parameters
    model.SetDblParam(COPT_DBLPARAM_TIMELIMIT, 10);

    // Solve the model
    model.Solve();

    // Analyze solution
    if (model.GetIntAttr(COPT_INTATTR_LPSTATUS) == COPT_LPSTATUS_OPTIMAL)
    {
      cout << endl;
      cout << "Objective value: " << model.GetDblAttr(COPT_DBLATTR_LPOBJVAL) << endl;

      VarArray vars = model.GetVars();
      cout << "Variable solution:" << endl;
      for (int i = 0; i < vars.Size(); ++i)
      {
        Var var = vars.GetVar(i);
        cout << "  " << var.GetName() << " = " << var.Get(COPT_DBLINFO_VALUE) << endl;
      }
      cout << endl;
    }
  }
  catch (CoptException e)
  {
    cout << "Error code = " << e.GetCode() << endl;
    cout << e.what() << endl;
  }
  catch (...)
  {
    cout << "Exception during optimization" << endl;
  }
}

void solve_mps(const char *szFileName)
{
  try
  {
    // Create COPT environment
    Envr env = Envr();

    // Create COPT model
    Model model = env.CreateModel("solve_mps");

    // Read SOCP from MPS file
    model.ReadMps(szFileName);

    // Set parameters
    model.SetDblParam(COPT_DBLPARAM_TIMELIMIT, 10);

    // Solve the model
    model.Solve();
  }
  catch (CoptException e)
  {
    cout << "Error code = " << e.GetCode() << endl;
    cout << e.what() << endl;
  }
  catch (...)
  {
    cout << "Exception during optimization" << endl;
  }
}

int main(int argc, char *argv[])
{
  // Solve SOCP with regular cone
  solve_soc();

  // Solve SOCP with rotated cone
  solve_rsoc();

  // Solve SOCP from file
  if (argc >= 2)
  {
    solve_mps(argv[1]);
  }

  return 0;
}
